#!/usr/bin/env python
########################################################################
# Copyright 2016,2021 VMware, Inc.  All rights reserved.
# -- VMware Confidential
########################################################################

"""Run any scripts generated by the kickstart mechanism

   The firstboot data is captured by the installer. Any scripts that need
   to be executed via the kickstart mechanism are added to a special
   directory.

   If UEFI secure boot is enabled, none of these scripts are executed.

"""

import os
import subprocess
import sys

from logging import (
   CRITICAL,
   DEBUG,
   FileHandler,
   Formatter,
   getLogger,
   Handler,
   INFO,
   StreamHandler,
)

scriptsDir = '/var/lib/vmware/firstboot'
loggerName = 'kickstartScript'
logFile = '/var/log/kickstart.log'

def setupLogger():
   """ Create and initialize a logger, ready for use.

   """

   logger = getLogger(loggerName)
   logger.setLevel(INFO)

   formatter = Formatter('%(asctime)s %(levelname)s %(message)s')

   streamHandler = StreamHandler()
   streamHandler.setFormatter(formatter)
   streamHandler.setLevel(INFO)
   logger.addHandler(streamHandler)

   try:
      fileHandler = FileHandler(logFile)
      fileHandler.setFormatter(formatter)
      fileHandler.setLevel(INFO)
      logger.addHandler(fileHandler)
   except Exception as e:
      logger.warning("Failed to set up log file %s: \"%s\"", logFile, e)

   return logger



def main(argv=None):

   logger = setupLogger()

   if len(os.listdir(scriptsDir)) == 0:
      # scriptsDir is empty, so no need to check secure boot status
      return

   from vmware.secureboot.secureBootCheck import IsSecureBootHost
   secureBootStatus = IsSecureBootHost()
   if secureBootStatus == 1:
      logger.info("SecureBoot is enabled, skipping execution of scripts in %s",
                  scriptsDir)
      return

   # Now that secureboot is not enabled, exec everything in the scriptsDir.

   for root, dirs, files in os.walk(scriptsDir):
      for f in files:
         path = os.path.join(root, f)
         if os.access(path, os.X_OK):
            logger.info("Running %s", path)
            try:
               output = subprocess.Popen(
                  path,
                  text=True,
                  stdout=subprocess.PIPE,
                  stderr=subprocess.STDOUT
               ).communicate()[0]
               logger.info(output)

            except Exception as e:
               logger.error("Failed to run %s: %s", path, e)

if __name__ == '__main__':
   sys.exit(main())
